@@ -25,15 +25,8 @@
 #endif
 
 #include <gst/gst.h>
-#include <gtk/gtkwidget.h>
-#include <gtk/gtktreeview.h>
-#include <gtk/gtkbutton.h>
-#include <gtk/gtkstock.h>
-#include <gtk/gtkrange.h>
-#include <gtk/gtkstatusbar.h>
-#include <gtk/gtkmessagedialog.h>
+#include <gtk/gtk.h>
 
-#include "bacon-volume.h"
 #include "sound-juicer.h"
 #include "sj-play.h"
 
@@ -43,6 +36,7 @@ static gint seek_to_track = -1, current_track = -1;
 static gboolean seeking = FALSE, internal_update = FALSE;
 static guint64 slen = GST_CLOCK_TIME_NONE;
 static gfloat vol = 1.0;
+static void set_gst_ui_and_play (void);
 
 static GtkTreeIter current_iter;
 
@@ -54,33 +48,9 @@ static GtkWidget *play_button, *next_menuitem, *prev_menuitem, *reread_menuitem,
 static gboolean
 select_track (void)
 {
+  GstStateChangeReturn ret;
   GstElement *cd;
 
-  if (seek_to_track == -1) {
-    gint tracks =
-        gtk_tree_model_iter_n_children (GTK_TREE_MODEL (track_store), NULL);
-
-    seek_to_track = 0;
-    while (seek_to_track < tracks) {
-      gboolean do_play;
-      GtkTreeIter next_iter;
-
-      if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
-                                          &next_iter, NULL, seek_to_track)) {
-        g_warning (G_STRLOC ": cannot get nth child");
-        return FALSE;
-      }
-      gtk_tree_model_get (GTK_TREE_MODEL (track_store), &next_iter,
-          COLUMN_EXTRACT, &do_play, -1);
-      if (do_play)
-        break;
-      seek_to_track++;
-    }
-    if (seek_to_track == tracks) {
-      seek_to_track = -1;
-      return FALSE;
-    }
-  }
   if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
                                      &current_iter, NULL, seek_to_track)) {
     g_warning (G_STRLOC ": cannot get nth child");
@@ -88,10 +58,24 @@ select_track (void)
   }
 
   cd = gst_bin_get_by_name_recurse_up (GST_BIN (pipeline), "cd-source");
-  gst_element_set_state (pipeline, GST_STATE_PAUSED);
-  gst_element_seek (pipeline, 1.0, gst_format_get_by_nick ("track"), GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, seek_to_track, GST_SEEK_TYPE_NONE, -1);
-  current_track = seek_to_track;
-  seek_to_track = -1;
+
+  ret = gst_element_set_state (pipeline, GST_STATE_PAUSED);
+  if (ret == GST_STATE_CHANGE_FAILURE) {
+    return FALSE;
+  } else if (ret == GST_STATE_CHANGE_SUCCESS) {
+    /* state change was instant, we can seek right away */
+    if (gst_element_seek (pipeline, 1.0, gst_format_get_by_nick ("track"),
+                          GST_SEEK_FLAG_FLUSH, GST_SEEK_TYPE_SET, seek_to_track,
+                          GST_SEEK_TYPE_NONE, -1)) {
+      current_track = seek_to_track;
+    } else {
+      /* seek failed - what now? */
+      return FALSE;
+    }
+  } else if (ret == GST_STATE_CHANGE_ASYNC) {
+    /* do nothing, seek will hopefully be done later?! */
+    current_track = seek_to_track;
+  }
 
   return TRUE;
 }
@@ -118,7 +102,7 @@ pause (void)
 }
 
 /**
- * Stop.
+ * Stop and reset UI.
  */
 static void
 stop (void)
@@ -126,11 +110,22 @@ stop (void)
   if (pipeline != NULL)
     gst_element_set_state (pipeline, GST_STATE_NULL);
 
+  /* TODO: this should be centralised into the state change logic really */
   gtk_widget_set_sensitive (next_menuitem, FALSE);
   gtk_widget_set_sensitive (prev_menuitem, FALSE);
   gtk_widget_set_sensitive (reread_menuitem, TRUE);
-  /* Call this to ensure that the UI is reset */
+  gtk_widget_hide (seek_scale);
+  gtk_widget_hide (volume_button);
   sj_main_set_title (NULL);
+  gtk_statusbar_push (GTK_STATUSBAR (statusbar), 0, "");
+  gtk_button_set_label (GTK_BUTTON (play_button), GTK_STOCK_MEDIA_PLAY);
+  slen = GST_CLOCK_TIME_NONE;
+
+  if (current_track != -1 &&
+      gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store), &current_iter, NULL, current_track))
+    gtk_list_store_set (track_store, &current_iter, COLUMN_STATE, STATE_IDLE, -1);
+
+  current_track = -1;
 }
 
 /**
@@ -142,6 +137,15 @@ is_playing (void)
   return (pipeline != NULL && GST_STATE (pipeline) == GST_STATE_PLAYING);
 }
 
+/**
+ * Are we paused?
+ */
+static gboolean
+is_paused (void)
+{
+  return (pipeline != NULL && GST_STATE (pipeline) == GST_STATE_PAUSED);
+}
+
 /*
  * callbacks.
  */
@@ -172,24 +176,8 @@ cb_hop_track (GstBus *bus, GstMessage *message, gpointer user_data)
     stop ();
     seek_to_track = 0;
   } else {
-    char *title;
     seek_to_track = next_track;
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_IDLE, -1);
-    select_track ();
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_PLAYING, -1);
-    gtk_tree_model_get (model,
-        &current_iter, COLUMN_TITLE, &title, -1);
-    sj_main_set_title (title);
-    g_free (title);
-
-    /* Set the Treelist focus to the item to be extracted */
-    GtkTreePath* path = gtk_tree_model_get_path(GTK_TREE_MODEL (track_store), &current_iter); 
-    gtk_tree_view_set_cursor (GTK_TREE_VIEW (track_listview), path, NULL, TRUE);
-    gtk_tree_path_free(path);
-
-    play ();
+    set_gst_ui_and_play ();
   }
 }
 
@@ -208,6 +196,11 @@ cb_error (GstBus *bus, GstMessage *message, gpointer user_data)
   gtk_widget_destroy (dialog);
   
   g_error_free (error);
+
+  /* There may be other (more generic) error messages on the bus; set pipeline
+   * to NULL state so these messages are flushed from the bus and we don't get
+   * called again for those */
+  stop ();
 }
 
 static gchar *
@@ -281,15 +274,15 @@ cb_state (GstBus *bus, GstMessage *message, gpointer user_data)
   GstState old_state, new_state;
   GstStateChange transition;
 
-  gst_message_parse_state_changed (message, &old_state, &new_state, NULL);
-  transition = GST_STATE_TRANSITION (old_state, new_state);
-
   /* all pipe elements will receive state transition messages,
    * so we filter those out. This won't be neccessary after
    * the playbin migration */
-  if ((GstElement*)GST_MESSAGE_SRC(message) != pipeline)
+  if ((GstElement*)GST_MESSAGE_SRC (message) != pipeline)
     return;
 
+  gst_message_parse_state_changed (message, &old_state, &new_state, NULL);
+  transition = GST_STATE_TRANSITION (old_state, new_state);
+
   if (transition == GST_STATE_CHANGE_READY_TO_PAUSED) {
     char *title;
     gtk_widget_show (seek_scale);
@@ -380,8 +373,12 @@ setup (GError **err)
     	              "read-speed", 2,
                       NULL);
     }
+    /* Disable paranoia in playback mode */
+    if (g_object_class_find_property (G_OBJECT_GET_CLASS (cdp), "paranoia-mode"))
+      g_object_set (cdp, "paranoia-mode", 0, NULL);
+
     g_object_set (G_OBJECT (cdp),
-                  "device", nautilus_burn_drive_get_device (drive),
+                  "device", brasero_drive_get_device (drive),
                   NULL);
 
     queue = gst_element_factory_make ("queue", "queue"); g_assert (queue);
@@ -444,32 +441,23 @@ on_play_activate (GtkWidget *button, gpointer user_data)
 
   if (is_playing ()) {
     pause ();
-  } else if (pipeline && GST_STATE (pipeline) == GST_STATE_PAUSED && 
-                (current_track == seek_to_track || seek_to_track == -1)) {
-    play ();
-  } else if (setup (&err)) {
-    char *title;
-    if (current_track != -1) {
-      if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
-                                          &current_iter, NULL, current_track))
-        return;
-    }
     gtk_list_store_set (track_store, &current_iter,
-                        COLUMN_STATE, STATE_IDLE, -1);
-    if (select_track ()) {
-      gtk_list_store_set (track_store, &current_iter,
-          COLUMN_STATE, STATE_PLAYING, -1);
-      gtk_tree_model_get (GTK_TREE_MODEL (track_store),
-          &current_iter, COLUMN_TITLE, &title, -1);
-      sj_main_set_title (title);
-      g_free (title);
-
-      play ();
-    } else {
-      g_warning (G_STRLOC ": failed to select track");
-      stop ();
-    }
-  } else {
+                        COLUMN_STATE, STATE_PAUSED, -1);
+ } else if (pipeline && GST_STATE (pipeline) == GST_STATE_PAUSED && 
+             current_track == seek_to_track) {
+    play ();
+    gtk_list_store_set (track_store, &current_iter,
+                        COLUMN_STATE, STATE_PLAYING, -1);
+  } else if (pipeline && GST_STATE (pipeline) == GST_STATE_PAUSED &&
+			 current_track != seek_to_track) {
+    if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
+                                          &current_iter, NULL, seek_to_track))
+        return;
+	set_gst_ui_and_play ();
+  } else if (setup (&err)) {
+	current_track = -1;
+	cb_hop_track (NULL,NULL,NULL);
+  }	else {
     GtkWidget *dialog;
 
     dialog = gtk_message_dialog_new (GTK_WINDOW (main_window), 0,
@@ -495,30 +483,9 @@ on_tracklist_row_activate (GtkTreeView * treeview, GtkTreePath * path,
   if (!gtk_tree_model_get_iter (model, &iter, path))
     return;
   gtk_tree_model_get (model, &iter, COLUMN_NUMBER, &track, -1);
-  if (gtk_list_store_iter_is_valid (GTK_LIST_STORE (model), &current_iter))
-      gtk_list_store_set (track_store, &current_iter, COLUMN_STATE, STATE_IDLE, -1);
-
   if (setup (&err)) {
-    char *title;
     seek_to_track = track - 1;
-
-    if (current_track != -1) {
-      if (!gtk_tree_model_iter_nth_child (GTK_TREE_MODEL (track_store),
-                                          &current_iter, NULL, current_track))
-        return;
-    }
-
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_IDLE, -1);
-    select_track ();
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_PLAYING, -1);
-    gtk_tree_model_get (GTK_TREE_MODEL (track_store),
-        &current_iter, COLUMN_TITLE, &title, -1);
-    sj_main_set_title (title);
-    g_free (title);
-    current_iter = iter;
-    play ();
+    set_gst_ui_and_play ();
   } else {
     GtkWidget *dialog;
 
@@ -564,18 +531,29 @@ on_previous_track_activate(GtkWidget *button, gpointer data)
     stop ();
     seek_to_track = 0;
   } else {
-    char *title;
     seek_to_track = prev_track;
-    gtk_list_store_set (track_store, &current_iter,
-        COLUMN_STATE, STATE_IDLE, -1);
-    select_track ();
+    set_gst_ui_and_play ();
+  }
+}
+
+static void
+set_gst_ui_and_play (void)
+{
+  char *title;
+
+  gtk_list_store_set (track_store, &current_iter,
+      COLUMN_STATE, STATE_IDLE, -1);
+  if (select_track ()) {
     gtk_list_store_set (track_store, &current_iter,
         COLUMN_STATE, STATE_PLAYING, -1);
-    gtk_tree_model_get (model,
+    gtk_tree_model_get (GTK_TREE_MODEL (track_store),
         &current_iter, COLUMN_TITLE, &title, -1);
     sj_main_set_title (title);
     g_free (title);
     play ();
+  } else {
+    g_warning (G_STRLOC ": failed to select track");
+    stop ();
   }
 }
 
@@ -586,7 +564,7 @@ on_tracklist_row_selected (GtkTreeView *treeview,
   gint track;
   GtkTreeSelection *selection = gtk_tree_view_get_selection (treeview);
 
-  if (is_playing ())
+  if (is_playing () || is_paused ())
     return;
 
   if (gtk_tree_selection_get_selected (selection, NULL, &current_iter)) {
@@ -601,9 +579,9 @@ on_tracklist_row_selected (GtkTreeView *treeview,
  */
 
 void
-on_volume_changed (GtkWidget * volb, gpointer data)
+on_volume_changed (GtkWidget * volb, gdouble value, gpointer data)
 {
-  vol = bacon_volume_button_get_value (BACON_VOLUME_BUTTON (volb));
+  vol = value;
   if (pipeline) {
     GstElement *volume;
 
